# -*- coding: utf-8 -*-
# vi:ts=4:et
#
# $Date: 2003/10/19 08:10:10 $
# $Revision: 1.6 $
# =====================================================

"""library for partition"""

# Return all the partitions of the number N.
# partition(3) -> (1,1,1), (2,1), (3,)

import sys
import itertools
import sets

# young
import mathformat
import error

__all__     = [
        'partition',
        'number_of_partition',
        ]

# cache partitions
_cache = {}


class Partition(tuple):
    """class for partition
    """
    def size(self):
        # XXX
        # do we need this?
        """Return the sum of parts of the partition
        """
        return sum(self)

    def get_num_of_parts(self):
        """Return the number of parts of the partition.
        """
        # If partition = (3,2,1,1),
        # we say that the partition has 4 parts.
        return len(self)

    def conjugate(self):
        """Return the conjugate of the partition.
        """
        # Stanley EC1 p.39

        # (4,3,1,1,1) -> (5,2,2,1)

        # * * * *
        # * * *
        # *
        # *
        # *
        # --> transpose
        # * * * * *
        # * *
        # * *
        # *

        length = len(self)
        padded_partition = self + (0,)
        buff = []

        for i in range(length, 0, -1):
            diff = padded_partition[i-1] - padded_partition[i]
            if diff > 0:
                buff.extend([i] * diff)

        return Partition(buff)

    def __repr__(self):
        return mathformat.pprint_partition(self)

    __str__ = __repr__

    def plot(self):
        """Plot the Ferrers diagram(Ferres graph).
        """
        # (3,1,1)
        # -->
        # ***
        # *
        # *
        for num in self:
            print "*" * num
        else:
            print

    def dot_notation(self):
        """Display partition with dots and bars.
        For example, partition (2,1,1) is "..|.|.".
        """
        buff = ["." * num for num in self]
        print "|".join(buff)

    def isdistinct(self):
        """Tests whether or not the partition consists of distinct parts.
        """
        return len(self) == len(sets.ImmutableSet(self))

class SeqOfPartition(object):
    """class for generating partitions
    """
    __slots__ = (
            'number',
            'seq',
            )

    def __init__(self, number):

        self.number = number
        self.seq    = []

        if number:
            self.init_partition()

    def __getitem__(self, index):
        return self.seq[index]

    def size(self):
        """Return the number of partitions
        """
        # Mathematica
        # Length[]

        # XXX
        # By convention we agree that p_0(0) = p(0) = 1.
        # (Stanley, "EC Vol.1", P.28.)

        if self.number == 0:
            return 1
        else:
            return len(self.seq)
    __len__ = size

    def get_partition(self):
        """Return the partition
        """
        return self.seq

    def add_partition(self, pat):
        """Add new partition
        """
        self.seq.append(pat)

    def __iter__(self):
        return iter(self.seq)

    def init_partition(self):
        """initialize partition.
        """
        try:
            self.seq = _cache[self.number]
        except KeyError, e:
            for i in range(1, self.number+1):
                try:
                    self.seq = _cache[i]
                except KeyError, e:
                    self.seq = _cache[i] = [Partition(pat) for pat in self.generate(i)]

    def plot(self):
        """Plot each partition
        """
        for pat in self.seq:
            pat.plot()

    def generate(self, n):
        """Partition generator
        """
        # trivial
        yield (n,)

        head, tail = n-1, 1

        while head > 0:
            if tail > 1:
                partial_result = _cache.get(tail, [])

                for candidate in partial_result[:]:
                    if _compare_max_num(head, candidate):
                        yield (head,) + candidate
            elif tail == 1:
                # tail == 1
                yield (head, tail)
            else:
                # tail == 0
                yield (head, )
            head, tail = head-1, tail+1

    def show_all_partition(self):
        """Show all the partitions in the cache
        """
        return show_partition(True)

    def show_partition(self, all = False):
        """Show the partition
        If all == False,
        show each partition of the number.
        If all == True,
        show all the partitions in the cache.
        """
        print self._show_partition(all)

    def _show_partition(self, all = False):
        buff = []
        if not all:
            buff = map(mathformat.pprint_partition, self.seq)
        else:

            for i in range(1, self.number+1):

                k, v = i, _cache[i]
                buff.append(`k`)
                for item in v:
                    buff.append("\t" + mathformat.pprint_partition(item))
        return "\n".join(buff)

    def dot_notation(self):
        """Display each partition in a dot-notation.
        """
        for part in self.seq:
            part.dot_notation()

    def distinct(self):
        self.seq = [part for part in itertools.ifilter(Partition.isdistinct, self.seq)]

    def __repr__(self):
        return self._show_partition()

    __str__ = __repr__


def _compare_max_num(number, part):
    """Tests whether or not the number is equal to or greater than
    any number of the partition."""
    # In short, if the partition is in order.
    # you only need to check if number >= part[0].

    # number = 3
    # part = (2,1)
    # --> OK

    # number = 2
    # part = (2,1,1,1)
    # --> OK

    # number = 3
    # part = (4,2,1)
    # --> NG

    return number >= part[0]


def partition(number):
    """Return the partition of number
    """
    # argument validity check
    if not (isinstance(number, (int, long)) and number >= 0):
        raise error.ArgumentError, ("number must be a non-negative integer.", number )

    return SeqOfPartition(number)

def number_of_partition(number, k=0):
    """Return the number of partitions of n.
    """
    if k:
        return partition_with_k_parts(number, k)
    else:
        return partition(number).size()

def partition_with_k_parts(n, k):
    """Return the number of partitions of n with exactly k parts.
    """
    # Stanley EC1 p.28

    # argument validity check
    if not (isinstance(n, (int, long)) and n >= 0):
        raise error.ArgumentError, ("number must be a non-negative integer.", n )

    if not (isinstance(k, (int, long)) and k >= 0):
        raise error.ArgumentError, ("number must be a non-negative integer.", k )


    if n == k:
        return 1
    elif k == 0 or n == 0:
        return 0
    elif n == k or k == 1:
        return 1
    elif n < k:
        return 0
    return partition_with_k_parts(n-1, k-1) + partition_with_k_parts(n - k, k)

